# -*- coding: utf-8 -*-

"""Interface to RDKit."""

import logging
import json
import pprint
import string

try:
    from rdkit import Chem
except ModuleNotFoundError:
    print(
        "Please install rdkit using conda:\n" "     conda install -c conda-forge rdkit"
    )
    raise

from seamm_util import CompactJSONEncoder

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# Valence
valence = {
    1: 1,
    2: 0,
    3: 1,
    4: 2,
    5: 3,
    6: 4,
    7: 3,
    8: 2,
    9: 1,
    10: 0,
    11: 1,
    12: 2,
    13: 3,
    14: 4,
    15: 3,
    16: 2,
    17: 1,
    18: 0,
    92: 2,
}

references = [
    """\
@misc{rdkit,
    title = {RDKit: Open-source cheminformatics.},
    month = {$month},
    year = {$year},
    author = {Greg Landrum, et al.},
    url = {https://www.rdkit.org},
    doi = {10.5281/zenodo.591637},
    version = {$version}
}
    """,
]


def rdkit_version():
    """Return the RDKit version."""
    return Chem.rdBase.rdkitVersion


def rdkit_citations():
    """Return the BibTeX citations for RDKit"""
    citations = []
    try:
        version = rdkit_version()
        tmp = version.split(".")
        year = tmp[0]
        if len(tmp) > 1:
            month = tmp[1]
        else:
            month = "?"

        for reference in references:
            template = string.Template(reference)
            citation = template.substitute(month=month, version=version, year=year)
            citations.append(citation)
    except Exception:
        pass
    return citations


class RDKitMixin:
    """A mixin for handling RDKit via its Python interface."""

    def coordinates_from_RDKMol(self, mol_or_conformer):
        """Update the coordinates from an RDKMol.

        Parameters
        ----------
        mol_or_conformer : rdkit.Chem.Mol or rdkit.Chem.rdchem.Conformer
            The RDKit molecule or conformer object
        """
        if isinstance(mol_or_conformer, Chem.Mol):
            if mol_or_conformer.GetNumConformers() == 0:
                raise ValueError("RDKit molecule has no conformers")
            conformer = mol_or_conformer.GetConformer()
        elif isinstance(mol_or_conformer, Chem.rdchem.Conformer):
            conformer = mol_or_conformer
        else:
            raise TypeError(
                f"Don't recognize molecule/conformer argument {type(mol_or_conformer)}"
            )
        self.atoms.coordinates = conformer.GetPositions()

    def coordinates_to_RDKMol(self, rdkmol):
        """Update the coordinates of an RDKMol from this configuration.

        Parameters
        ----------
        rdkmol : rdkit.Chem.RWMol
            The RDKMol to use.
        """
        rdkconf = rdkmol.GetConformers()[0]
        rdkconf.SetPositions(self.atoms.coordinates)

    def to_RDKMol(self, properties=None):
        """Return an RDKMol object for the configuration, template, or subset."""
        index = {}
        indices = []
        rdk_mol = Chem.RWMol()
        for atno, _id in zip(self.atoms.atomic_numbers, self.atoms.ids):
            atom = Chem.Atom(atno)
            idx = rdk_mol.AddAtom(atom)
            index[_id] = idx
            indices.append(idx)

        bond_types = {
            1: Chem.BondType.SINGLE,
            2: Chem.BondType.DOUBLE,
            3: Chem.BondType.TRIPLE,
            5: Chem.BondType.AROMATIC,
        }
        for row in self.bonds.bonds():
            if row["bondorder"] > 0:
                rdk_mol.AddBond(
                    index[row["i"]], index[row["j"]], bond_types[row["bondorder"]]
                )

        # Check for NH4+ type groups and set their charge
        rdk_mol.UpdatePropertyCache(strict=False)
        for at in rdk_mol.GetAtoms():
            atno = at.GetAtomicNum()
            if atno in valence:
                if atno == 5:
                    # Boron is a special case! Both BH4- and BH2- have a - charge.
                    if at.GetExplicitValence() == 4:
                        charge = -1
                    elif at.GetExplicitValence() == 2:
                        charge = -1
                    else:
                        charge = 0
                else:
                    charge = at.GetExplicitValence() - valence[atno]
                if charge != 0 and at.GetFormalCharge() == 0:
                    at.SetFormalCharge(charge)

        natom = self.atoms.n_atoms
        conf = Chem.Conformer(natom)
        for idx, xyz in zip(indices, self.atoms.coordinates):
            conf.SetAtomPosition(idx, xyz)

        rdk_mol.AddConformer(conf)
        try:
            Chem.SanitizeMol(rdk_mol)
        except Chem.KekulizeException as e:
            logger.warning(f"Kekulization failed: {e}")
            flags = (
                Chem.SanitizeFlags.SANITIZE_ALL ^ Chem.SanitizeFlags.SANITIZE_KEKULIZE
            )
            Chem.rdmolops.SanitizeMol(rdk_mol, sanitizeOps=flags)

        # Add the net charge and spin multiplicity as properties for configurations
        if self.__class__.__name__ == "_Configuration":
            rdk_mol.SetIntProp("SEAMM|net charge|int|", self.charge)
            rdk_mol.SetIntProp("SEAMM|spin multiplicity|int|", self.spin_multiplicity)
            rdk_mol.SetProp(
                "SEAMM|XYZ|json|",
                json.dumps(self.coordinates, indent=4, cls=CompactJSONEncoder),
            )
            if self.system.name is not None:
                rdk_mol.SetProp("SEAMM|system name|str|", self.system.name)
            if self.name is not None:
                rdk_mol.SetProp("SEAMM|configuration name|str|", self.name)

        if properties is not None:
            data = self.properties.get(properties, include_system_properties=True)
            for _property, value in data.items():
                _type = self.properties.type(_property)
                units = self.properties.units(_property)
                value = value["value"]
                key = f"SEAMM|{_property}|{_type}|"
                if units is not None and units != "":
                    key += units

                if _type == "int":
                    rdk_mol.SetIntProp(key, value)
                elif _type == "float":
                    rdk_mol.SetDoubleProp(key, value)
                elif _type == "json":
                    rdk_mol.SetProp(key, json.dumps(value))
                else:
                    rdk_mol.SetProp(key, str(value))

        return rdk_mol

    def from_RDKMol(
        self,
        mol_or_conformer,
        properties="all",
        atoms=True,
        coordinates=True,
        bonds=True,
    ):
        """Transform an RDKit molecule into the current object.

        Parameters
        ----------
        mol_or_conformer : rdkit.Chem.Mol or rdkit.Chem.rdchem.Conformer
            The RDKit molecule or conformer object

        properties : str = "all"
            Whether to include all properties or none

        atoms : bool = True
            Recreate the atoms

        coordinates : bool = True
            Update the coordinates

        bonds : bool = True
            Recreate the bonds from the RDKit molecule

        Returns
        -------
        (str, str) : system_name, configuration_name
        """
        system_name = None
        configuration_name = None

        if isinstance(mol_or_conformer, Chem.Mol):
            mol = mol_or_conformer
            if mol.GetNumConformers() == 0:
                raise ValueError("RDKit molecule has no conformers")
            conformer = mol.GetConformer()
        elif isinstance(mol_or_conformer, Chem.rdchem.Conformer):
            conformer = mol_or_conformer
            mol = conformer.GetOwningMol()
        else:
            raise RuntimeError(
                f"Don't recognize molecule/conformer argument {type(mol_or_conformer)}"
            )

        atnos = []
        for rdk_atom in mol.GetAtoms():
            atnos.append(rdk_atom.GetAtomicNum())
            logger.debug(f"atom {atnos}")

        Xs = []
        Ys = []
        Zs = []
        for atom_coords in conformer.GetPositions():
            Xs.append(atom_coords[0])
            Ys.append(atom_coords[1])
            Zs.append(atom_coords[2])
            logger.debug(f"{atom_coords[0]} {atom_coords[1]} {atom_coords[2]}")

        Is = []
        Js = []
        BondOrders = []
        bond_types = {
            Chem.BondType.SINGLE: 1,
            Chem.BondType.DOUBLE: 2,
            Chem.BondType.TRIPLE: 3,
            Chem.BondType.AROMATIC: 5,
        }
        for rdk_bond in mol.GetBonds():
            i = rdk_bond.GetBeginAtom().GetIdx()
            j = rdk_bond.GetEndAtom().GetIdx()
            bondorder = bond_types[rdk_bond.GetBondType()]
            Is.append(i)
            Js.append(j)
            BondOrders.append(bondorder)
            logger.debug(f"bond {i} - {j} {bondorder}")

        if atoms:
            self.clear()
            try:
                ids = self.atoms.append(x=Xs, y=Ys, z=Zs, atno=atnos)
            except Exception:
                print("Xs=")
                pprint.pp(Xs)
                print("Ys=")
                pprint.pp(Ys)
                print("Zs=")
                pprint.pp(Zs)
                print("atnos=")
                pprint.pp(atnos)
                raise
        else:
            ids = self.atoms.ids

            if coordinates:
                xyz = [[x, y, z] for x, y, z in zip(Xs, Ys, Zs)]
                self.atoms.coordinates = xyz

        if atoms or bonds:
            i = [ids[x] for x in Is]
            j = [ids[x] for x in Js]
            self.bonds.append(i=i, j=j, bondorder=BondOrders)

        data = mol.GetPropsAsDict()
        if self.__class__.__name__ == "_Configuration":
            # Charge
            self.charge = int(Chem.GetFormalCharge(mol))

            # Calculate spin multiplicity assuming maximal spin
            n_electrons = 0
            for rdk_atom in mol.GetAtoms():
                n_electrons += rdk_atom.GetNumRadicalElectrons()
            self.spin_multiplicity = n_electrons + 1

            # Check for property items for charge and multiplicity
            if "SEAMM|net charge|int|" in data:
                self.charge = int(data["SEAMM|net charge|int|"])
                del data["SEAMM|net charge|int|"]
            if "SEAMM|spin multiplicity" in data:
                self.spin_multiplicity = int(data["SEAMM|spin multiplicity|int|"])
                del data["SEAMM|spin multiplicity|int|"]

            # Coordinates
            if coordinates and "SEAMM|XYZ|json|" in data:
                self.coordinates = json.loads(data["SEAMM|XYZ|json|"])
                del data["SEAMM|XYZ|json|"]

            if "SEAMM|system name|str|" in data:
                system_name = data["SEAMM|system name|str|"]
                del data["SEAMM|system name|str|"]
            if "SEAMM|configuration name|str|" in data:
                configuration_name = data["SEAMM|configuration name|str|"]
                del data["SEAMM|configuration name|str|"]

        # Record any properties in the database if desired
        if properties == "all":
            for key, value in data.items():
                if key.startswith("SEAMM|"):
                    _, _property, _type, units = key.split("|", 4)
                    units = None if units.strip() == "" else units
                    if not self.properties.exists(_property):
                        self.properties.add(_property, _type=_type, units=units)

                    if _type == "json":
                        value = json.dumps(value)

                    self.properties.put(_property, value)
                else:
                    if not self.properties.exists(key):
                        _type = value.__class__.__name__
                        self.properties.add(key, _type)

                    self.properties.put(key, value)
        return system_name, configuration_name

    def get_SSSR(self, symmetrized=False):
        """Calculate the smallest set of smallest rings.

        Parameters
        ----------
        symmetrized : bool = False
            Whether to return the symmetrized SSSR

        Returns
        -------
        list of list of int
        """
        if symmetrized:
            return Chem.rdmolops.GetSymSSSR(self.to_RDKMol())

        return Chem.rdmolops.GetSSSR(self.to_RDKMol())



# -*- coding: utf-8 -*-

"""Functions for handling PubChem"""

import logging
import re
from time import perf_counter_ns, sleep
from urllib.parse import quote as url_quote

import requests

logger = logging.getLogger(__name__)
pug_url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug"

try:
    import pubchempy as pcp
except ModuleNotFoundError:
    print(
        "To use PubChem, please install pubchempy using conda:\n"
        "     conda install -c conda-forge pubchempy"
    )
    logger.warning(
        "To use PubChem, please install pubchempy using conda:\n"
        "     conda install -c conda-forge pubchempy"
    )


def PC_standardize(structures):
    """Use the PubChem standardization service with SMILES.

    Parameters
    ----------
    structures : [str]
        The SMILES of the structures to standardize

    Returns
    -------
    [{str: str}]
        A list of dicts of the results.

    Note:
        The response contains information to throttle the requests, like this:

        X-Throttling-Control: Request Count status: Green (0%),
                              Request Time status: Green (0%),
                              Service status: Green (20%)
    """
    results = []
    max_count = 0
    max_time = 0
    max_service = 0
    n_sleep = 0
    with requests.Session() as session:
        t0 = perf_counter_ns()
        for SMILES in structures:
            while True:
                response = session.post(
                    f"{pug_url}/standardize/SMILES/JSON",
                    data={"smiles": SMILES, "include_components": False},
                )

                # Check the throttling request
                if "X-Throttling-Control" in response.headers:
                    throttling_header = response.headers["X-Throttling-Control"]
                    match = re.search(
                        r"Request Count status: \w+ \((\d+)%\), "
                        r"Request Time status: \w+ \((\d+)%\), "
                        r"Service status: \w+ \((\d+)%\)",
                        throttling_header,
                    )
                    if match:
                        count = int(match.group(1))
                        time = int(match.group(2))
                        service = int(match.group(3))

                        max_count = max(count, max_count)
                        max_time = max(time, max_time)
                        max_service = max(service, max_service)

                        if count > 75 or time > 75 or service > 75:
                            n_sleep += 1
                            sleep(1)

                if response.status_code == 503:
                    print("status code = 503")
                    continue

                response.raise_for_status()

                break

            data = response.json()

            # Pull out some info from the data
            cmpds = data["PC_Compounds"]
            if len(cmpds) > 1:
                raise ValueError(
                    f"There are {len(cmpds)} in the PubChem standardization!"
                )

            cmpd = cmpds[0]
            result = {"original SMILES": SMILES}
            if "id" in cmpd and "id" in cmpd["id"] and "cid" in cmpd["id"]["id"]:
                result["cid"] = cmpd["id"]["id"]["cid"]

            for props in cmpd["props"]:
                if "urn" in props and "value" in props and "label" in props["urn"]:
                    key = props["urn"]["label"]
                    value = props["value"]
                    _type = [*value.keys()][0]
                    if _type == "sval":
                        result[key] = value[_type]
                    else:
                        raise TypeError(f"Unknown type {_type} in compound properties.")

            results.append(result)
        t1 = perf_counter_ns()
        t = (t1 - t0) / 1000000000
        n = len(structures)
        per = t / n

    return {
        "data": results,
        "n_structures": len(structures),
        "throttling": {
            "maximum count": max_count,
            "maximum time": max_time,
            "maximum service": max_service,
            "number of slowdowns": n_sleep,
        },
        "time": {
            "total": t,
            "per structure": per,
        },
    }


class PubChemMixin:
    """A mixin for handling the PubChem database."""

    @property
    def PC_cid(self):
        """Return the PubChem CID for this structure, or None."""
        inchi = self.to_inchi()
        results = pcp.get_compounds(inchi, "inchi")

        if len(results) == 0:
            return None

        if len(results) >= 1:
            logger.info(f"PubChem search returned more than one hit of {inchi}")

        result = results[0].to_dict()
        if "cid" in result:
            return result["cid"]

        return None

    def PC_from_cid(self, cid, fallback=None):
        """Create the configuration from the PubChem 3-D structure, if available.

        Parameters
        ----------
        cid : int
            The PubChem CID.

        fallback : str
            A fallback SMILES, InChI, etc. to use if PubChem fails
        """
        response = requests.get(f"{pug_url}/compound/cid/{cid}/SDF")
        if response.status_code == 200:
            self.from_sdf_text(response.text)
            return

        # An error!
        if fallback is None:
            raise RuntimeError(f"No 3-D structure available for {cid}")

        # See what the fallback is
        if len(fallback) == 27:
            tmp = fallback.split("-")
            if (
                len(tmp) == 3
                and len(tmp[0]) == 14
                and len(tmp[1]) == 10
                and len(tmp[2]) == 1
            ):
                self.from_inchikey(fallback)
                return
        if fallback[0:7] == "InChI=":
            self.from_inchi(fallback)
            return
        else:
            self.from_smiles(fallback)
            return

    def PC_from_identifier(
        self, identifier, namespace="detect", fallback=None, properties="all"
    ):
        """Create the configuration from the PubChem 3-D structure, if available.

        Parameters
        ----------
        identifier : int or str
            The PubChem identifier
        namespace : str
            The PubChem namespace: cid, name, smiles, inchi, inchikey
        fallback : str
            A fallback SMILES, InChI, etc. to use if PubChem fails
        properties : str = "all"
            Whether to include all properties or none
        """
        if namespace == "detect":
            # Work through the possibilities
            if len(identifier) == 27:
                tmp = identifier.split("-")
                if (
                    len(tmp) == 3
                    and len(tmp[0]) == 14
                    and len(tmp[1]) == 10
                    and len(tmp[2]) == 1
                ):
                    namespace = ["inchikey"]
            elif identifier[0:7] == "InChI=":
                self.from_inchi(identifier)
                namespaces = ["inchi"]
            else:
                namespaces = ["name", "smiles"]
        else:
            namespaces = [namespace]

        for namespace in namespaces:
            response = requests.get(
                f"{pug_url}/compound/{namespace}/{url_quote(identifier)}/"
                "SDF?record_type=3d"
            )
            if response.status_code == 200:
                self.from_sdf_text(response.text, properties=properties)
                return

        # An error!
        if fallback is None:
            raise RuntimeError(f"No 3-D structure available for {identifier}")

        # See what the fallback is
        if len(fallback) == 27:
            tmp = fallback.split("-")
            if (
                len(tmp) == 3
                and len(tmp[0]) == 14
                and len(tmp[1]) == 10
                and len(tmp[2]) == 1
            ):
                self.from_inchikey(fallback)
                return
        if fallback[0:7] == "InChI=":
            self.from_inchi(fallback)
            return
        else:
            self.from_smiles(fallback)
            return

    def PC_iupac_name(self, fallback=None):
        """Return the IUPAC name for this structure, or None.

        Parameters
        ----------
        fallback : str
            A name to return if PubChem doesn't have a name
        """
        inchi = self.to_inchi()
        results = pcp.get_compounds(inchi, "inchi")

        if len(results) == 0:
            return fallback

        if len(results) >= 1:
            logger.info(f"PubChem search returned more than one hit of {inchi}")

        result = results[0].to_dict()
        if "iupac_name" in result:
            return result["iupac_name"]

        return fallback



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the QCSchema mixin of the class."""

import json
import pprint  # noqa: F401
import pytest  # noqa: F401


def test_water(H2O):
    """Test creating QCSchema for water."""
    correct = {
        "schema_name": "qcschema_molecule",
        "schema_version": 2,
        "symbols": ["O", "H", "H"],
        "geometry": [0.0, 0.0, 0.0, 1.430429, 0.0, 1.107157, -1.430429, 0.0, 1.107157],
        "molecular_charge": 0,
        "molecular_multiplicity": 1,
        "connectivity": [[0, 1, 1], [0, 2, 1]],
        "fragments": [[0, 1, 2]],
        "name": "water / TIP3P",
        "atom_labels": ["O", "H1", "H2"],
    }

    data = H2O.to_qcschema_json()

    result = json.loads(data)
    result["geometry"] = [round(v, 6) for v in result["geometry"]]
    for key in correct.keys():
        if key not in result:
            print("----")
            pprint.pp(result)
            print("----")

            assert False, f"{key} missing from result"
        elif correct[key] != result[key]:
            print("----")
            pprint.pp(result)
            print("----")

            print(f"values for {key} differ!")
            assert correct[key] == result[key]

    for key in result.keys():
        if key not in correct:
            print("----")
            pprint.pp(result)
            print("----")

            assert False, f"extra key '{key}' in result"


def test_acetic_acid(AceticAcid):
    """Test creating QCSchema for actic acid."""
    correct = {
        "schema_name": "qcschema_molecule",
        "schema_version": 2,
        "symbols": ["C", "H", "H", "H", "C", "O", "O", "H"],
        "geometry": [
            2.040337,
            0.034204,
            -0.034771,
            1.09264,
            5.929205,
            0.53158,
            1.362304,
            -1.27292,
            -1.485136,
            1.332635,
            -0.593941,
            1.80072,
            1.079601,
            2.62653,
            -0.597342,
            -0.250011,
            3.239369,
            -2.375008,
            1.843806,
            4.340701,
            1.118529,
            4.105241,
            0.030425,
            -0.057826,
        ],
        "molecular_charge": 0,
        "molecular_multiplicity": 1,
        "connectivity": [
            [0, 1, 1],
            [0, 2, 1],
            [0, 3, 1],
            [0, 4, 1],
            [4, 5, 2],
            [4, 6, 1],
            [6, 7, 1],
        ],
        "fragments": [[0, 1, 2, 3, 4, 5, 6, 7]],
        "name": "acetic acid / acetic acid",
    }

    data = AceticAcid.to_qcschema_json()

    result = json.loads(data)
    result["geometry"] = [round(v, 6) for v in result["geometry"]]
    for key in correct.keys():
        if key not in result:
            print("----")
            pprint.pp(result)
            print("----")

            assert False, f"{key} missing from result"
        elif correct[key] != result[key]:
            print("----")
            pprint.pp(result)
            print("----")

            print(f"values for {key} differ!")
            assert correct[key] == result[key]

    for key in result.keys():
        if key not in correct:
            print("----")
            pprint.pp(result)
            print("----")

            assert False, f"extra key '{key}' in result"


def test_from_schema(configuration, AceticAcid):
    """Test creating QCSchema for actic acid."""
    correct = '{"schema_name": "qcschema_molecule", "schema_version": 2, "symbols": ["C", "H", "H", "H", "C", "O", "O", "H"], "geometry": [2.040337296754675, 0.03420404285566326, -0.03477096069304994, 1.092639645256602, 5.929204688614864, 0.5315799588562472, 1.3623035632402012, -1.2729195175455674, -1.4851357613406495, 1.3326348630836315, -0.5939409209687825, 1.80072002415257, 1.0796005349967084, 2.626530340612506, -0.5973424279931026, -0.25001076628752755, 3.2393685228275113, -2.3750077934252807, 1.843805779793958, 4.340700908257376, 1.1185288931639272, 4.10524103312944, 0.030424590606418698, -0.05782561941344175], "molecular_charge": 0, "molecular_multiplicity": 1, "connectivity": [[0, 1, 1], [0, 2, 1], [0, 3, 1], [0, 4, 1], [4, 5, 2], [4, 6, 1], [6, 7, 1]], "fragments": [[0, 1, 2, 3, 4, 5, 6, 7]], "name": "acetic acid / acetic acid"}'  # noqa: E501

    configuration.from_qcschema_json(correct)

    assert configuration == AceticAcid



# -*- coding: utf-8 -*-

"""Alignment methods for configurations"""

import logging
from math import sqrt

import numpy as np

try:
    import openbabel  # noqa: F401
    import openbabel.openbabel as ob
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "conda install -c conda-forge openbabel"
    )
    raise

try:
    import rdkit
    from rdkit import Chem
    from rdkit.Chem import rdMolAlign, rdmolops, AllChem
except ModuleNotFoundError:
    print("Please install rdkit using conda:\n     conda install -c conda-forge rdkit")
    raise

logger = logging.getLogger(__name__)


def RMSD(
    structure, reference, include_h=False, symmetry=False, align=False, flavor=None
):
    """Compute the RMSD between configurations.

    Parameters
    ----------
    structure : Configuration, RDKMol, or OBMol

    reference : Configuration, RDKMol, or OBMol
        A single configuration to match. if 'structure' is not a configuration,
        this must match the type of 'structure'.

    include_h : bool = True
        Whether to include hydrogen atoms in the RMSD

    symmetry : bool = False
        Whether to detect symmetric flips. Note this requires a lot of memory for
        larger systems.

    align : bool = False
        Whether to transform the structure to the best alignment with
        the reference.

    flavor : str = None
        The flavor to use. Currently 'rdkit' or 'openbabel'

    Returns
    -------
    {str: float|int}
        RMSD: The RMS between the structure and the reference.
        maximum displacement: The maximum displacement of any atom.
        displaced atom: The index of the atom with the maximum displacement.
    """
    from .configuration import _Configuration

    if isinstance(structure, _Configuration):
        if isinstance(reference, _Configuration):
            if flavor is None or flavor == "rdkit":
                _structure = structure.to_RDKMol()
                _reference = reference.to_RDKMol()
            elif flavor == "openbabel":
                _structure = structure.to_OBMol()
                _reference = reference.to_OBMol()
            else:
                raise ValueError(f"Cannot handle flavor '{flavor}'")
        elif isinstance(reference, Chem.RWMol):
            if flavor is None or flavor == "rdkit":
                _structure = structure.to_RDKMol()
                _reference = Chem.RWMol(reference)
            else:
                raise ValueError(
                    f"Cannot use flavor '{flavor}' with an RDKMol reference"
                )
        elif isinstance(reference, ob.OBMol):
            if flavor is None or flavor == "openbabel":
                _structure = structure.to_OBMol()
                _reference = ob.OBMol(reference)
            else:
                raise ValueError(
                    f"Cannot use flavor '{flavor}' with an OBMol reference"
                )
        else:
            raise ValueError(
                "Reference must be a Configuration, RDKMol, or OBMol, not "
                f"{type(reference)}"
            )
    elif isinstance(structure, rdkit.Chem.rdchem.RWMol):
        if flavor is not None and flavor != "rdkit":
            raise ValueError(f"Cannot use flavor '{flavor}' with an RDKMol structure")
        if isinstance(reference, _Configuration):
            _structure = Chem.rdchem.RWMol(structure)
            _reference = reference.to_RDKMol()
        elif isinstance(reference, rdkit.Chem.rdchem.RWMol):
            _structure = Chem.rdchem.RWMol(structure)
            _reference = Chem.rdchem.RWMol(reference)
        elif isinstance(reference, ob.OBMol):
            raise ValueError(
                "If the structure is an RDKMol, the reference "
                f"must be a Configuration or RDKMol, not {type(reference)}"
            )
        else:
            raise ValueError(
                "Reference must be a Configuration, RDKMol, or OBMol, not "
                f"{type(reference)}"
            )
    elif isinstance(structure, ob.OBMol):
        if flavor is not None and flavor != "openbabel":
            raise ValueError(f"Cannot use flavor '{flavor}' with an OBMol structure")
        if isinstance(reference, _Configuration):
            _structure = ob.OBMol(structure)
            _reference = reference.to_OBMol()
        elif isinstance(reference, rdkit.Chem.rdchem.RWMol):
            raise ValueError(
                "If the structure is an OBMol, the reference "
                f"must be a Configuration or OBMol, not {type(reference)}"
            )
        elif isinstance(reference, ob.OBMol):
            _structure = ob.OBMol(structure)
            _reference = ob.OBMol(reference)
        else:
            raise ValueError(
                f"Reference must be a Configuration, or OBMol, not {type(reference)}"
            )
    else:
        raise ValueError(
            "Structure must be a Configuration, RDKMol, or OBMol, "
            f"not {type(structure)}"
        )

    if isinstance(_structure, rdkit.Chem.rdchem.RWMol):
        result = RDK_RMSD(
            _structure,
            _reference,
            include_h=include_h,
            symmetry=symmetry,
        )
        if align:
            if isinstance(structure, _Configuration):
                structure.coordinates_from_RDKMol(_structure)
            else:
                structure.GetConformer(0).SetPositions(
                    _structure.GetConformer(0).GetPositions()
                )
    elif isinstance(_structure, ob.OBMol):
        result = OB_RMSD(
            _structure,
            _reference,
            include_h=include_h,
            symmetry=symmetry,
        )
        if align:
            if isinstance(structure, _Configuration):
                structure.coordinates_from_OBMol(_structure)
            else:
                for at1, at2 in zip(
                    ob.OBMolAtomIter(structure), ob.OBMolAtomIter(_structure)
                ):
                    at1.SetVector(at2.GetVector())

    return result


def RDK_RMSD(structure, reference, include_h=True, symmetry=False):
    """Compute the RMSD between configurations using RDKit.

    Parameters
    ----------
    structure: RDKMol
        The structure to align

    reference : RDKMol
        The structure to compare to.

    include_h : bool = True
        Whether to include hydrogen atoms in the RMSD

    symmetry : bool = False
        Whether to detect symmetric flips. Note this requires a lot of memory for
        larger systems.

    Returns
    -------
    {str: float|int}
        RMSD: The RMS between the structure and the reference.
        maximum displacement: The maximum displacement of any atom.
        displaced atom: The index of the atom with the maximum displacement.

    Note
    ----
    The structure is modified in place to be optimally aligned with the
    reference.
    """
    if not include_h:
        structure = rdmolops.RemoveHs(structure)
        reference = rdmolops.RemoveHs(reference)

    if symmetry:
        rmsd, transform, atom_map = rdMolAlign.GetBestAlignmentTransform(
            structure, reference
        )
        AllChem.TransformMol(structure, transform)

        max_displacement = -1.0
        displaced_atom = None
        XYZ1 = structure.GetConformer(0).GetPositions()
        XYZ2 = reference.GetConformer(0).GetPositions()
        for at1, at2 in atom_map:
            dist = np.linalg.norm(XYZ1[at1] - XYZ2[at2])
            if dist > max_displacement:
                max_displacement = float(dist)
                displaced_atom = at1
    else:
        rmsd, transform = rdMolAlign.GetAlignmentTransform(structure, reference)
        AllChem.TransformMol(structure, transform)

        max_displacement = 0.0
        displaced_atom = None
        XYZ1 = structure.GetConformer(0).GetPositions()
        XYZ2 = reference.GetConformer(0).GetPositions()
        for count, XYZ1_XYZ2 in enumerate(zip(XYZ1, XYZ2)):
            XYZ1, XYZ2 = XYZ1_XYZ2
            dist = np.linalg.norm(XYZ1 - XYZ2)
            if dist > max_displacement:
                max_displacement = float(dist)
                displaced_atom = count
    return {
        "RMSD": rmsd,
        "maximum displacement": max_displacement,
        "displaced atom": displaced_atom,
    }


def OB_RMSD(structure, reference, include_h=True, symmetry=False):
    """Compute the RMSD between configurations using OpenBabel.

    Parameters
    ----------
    structure: OBMol
        The structure to align

    reference : OBMol
        The structure to compare to.

    include_h : bool = True
        Whether to include hydrogen atoms in the RMSD

    symmetry : bool = False
        Whether to detect symmetric flips. Note this requires a lot of memory for
        larger systems.

    Returns
    -------
    {str: float|int}
        RMSD: The RMS between the structure and the reference.
        maximum displacement: The maximum displacement of any atom.
        displaced atom: The index of the atom with the maximum displacement.

    Note
    ----
    The structure is modified in place to be optimally aligned with the
    reference.
    """
    align = ob.OBAlign(include_h, symmetry)
    align.SetTargetMol(structure)

    try:
        align.SetRefMol(reference)
        align.Align()
        align.UpdateCoords(structure)
        rmsd = align.GetRMSD()

        max_displacement = -1.0
        displaced_atom = None
        for count, at1_at2 in enumerate(
            zip(ob.OBMolAtomIter(structure), ob.OBMolAtomIter(reference))
        ):
            at1, at2 = at1_at2
            dist = sqrt(
                (at1.x() - at2.x()) ** 2
                + (at1.y() - at2.y()) ** 2
                + (at1.z() - at2.z()) ** 2
            )
            if dist > max_displacement:
                max_displacement = float(dist)
                displaced_atom = count
    except TypeError:
        raise
    except Exception as e:
        logger.error(f"Configuration.RMSD {e}")
        raise
    return {
        "RMSD": rmsd,
        "maximum displacement": max_displacement,
        "displaced atom": displaced_atom,
    }


class AlignMixin:
    """A mixin for handling alignment of configuration."""

    def RMSD(
        self, reference, include_h=False, symmetry=False, align=False, flavor="rdkit"
    ):
        """Compute the RMSD between configurations.

        Parameters
        ----------
        reference : Configuration|OBMol|RDKMol
            The structure to compare and possibly align to

        include_h : bool = True
            Whether to include hydrogen atoms in the RMSD

        symmetry : bool = False
            Whether to detect symmetric flips. Note this requires a lot of memory for
            larger systems.

        align : bool = False
            Whether to transform the structure to the best alignment with
            the target.

        flavor : str = "rdkit"
            The flavor to use. Currently 'rdkit' or 'openbabel'

        Returns
        -------
        float or [float]
            The RMS between the structure configuration and the target(s).
        """
        return RMSD(
            self,
            reference,
            include_h=include_h,
            symmetry=symmetry,
            align=align,
            flavor=flavor,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `molsystem` package.

What is the version of SQLITE?
"""

import sqlite3


def test_sqlite_version():
    """Print the version of SQLite"""
    print(f"SQLite version {sqlite3.sqlite_version}")
    assert True
